//#define EBATWO
//#define ADDON2
#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using System.Reflection;
using MVR.FileManagementSecure;
#endregion

namespace Cotyounoyume{
	public class EBATCore : MVRScript
	{
		public float version = 1.33f;

		#region common parameters
		public Atom atom;
		public SimpleJSON.JSONClass settingsRoot = new SimpleJSON.JSONClass();
		public SimpleJSON.JSONClass settings = new SimpleJSON.JSONClass();
		public SimpleJSON.JSONClass sceneSettings = new SimpleJSON.JSONClass();
		public string myPath;
		public bool maintenanceAlert = false;
		public float lastTouchingTime = 0f;
		public string gender;
		public bool orgasmModeOn = false;
		public bool addon2ModeOn = false;
		public float perSecondOffset = 1f;
		public float lastTime = 0f;
		
		protected float openingTime;
		public float currentTime;
		public int gauge = 0;
		public float calmingDownTime;
		public float calmingDownTimeOffset = 2;
		public float calmingDownTimeMinusOffset;
		public int calmingDownPlus;
		public bool isHot = false;
		
		public int curStage = -1;
		public int preStage = -1;
		protected string defaultSettingsPath = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/Config/defaultSettings.json";
		protected string currentSettingsPath = "Custom/Scripts/ExpressionBlushingAndTears_currentSettings.json";
		protected string myClassName = "EBATCore";

		public string pluginLabel;
		public bool disableLog = true;
		public int[] stageThreshold;
		public bool isBreathActive = false;
		public bool isDuringOrgasm = false;
		public bool isDuringAfterglow = false;
		public string myName = "";
		
		public Expression expression;
		public Blushing blushing;
		public Tears tears;
		public EyesControl eyesControl;
		public Belly belly;
		public Etc etc;
		public BodyMorph bodyMorph;
		public UnityAssets unityAssets;
		public WorldUI worldUi;
		public Sound sound;
		public CustomUI customUi;
		
		public bool menuOnGlobal = true;
		public List<Atom> generatedAtoms = new List<Atom>();
		public Dictionary<string, Atom> generatedAtomsDictionary = new Dictionary<string, Atom>();
		public AddonManager1 addonManager1;

#if	(EBATWO)
		public Orgasm orgasm;
#endif
#if	(ADDON2)
		public AddonManager2 addonManager2;
#endif
		#endregion
				
		public override void Init(){
			try{
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("This plug-in have to be added to person type atom.");
                    return;
                }
				atom = containingAtom;
#if	(EBATWO)
				orgasmModeOn = true;
#endif
#if	(ADDON2)
				addon2ModeOn = true;
#endif
				
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				gender = character.gender.ToString();
				
				myName = atom.name;
				logger("atom name: " + myName);
				myPath = Utility.GetPluginPath(this as MVRScript, atom);
				logger("myPath: " + myPath);
				
				expression  = new Expression(this);
				blushing    = new Blushing(this);
				tears       = new Tears(this);
				bodyMorph   = new BodyMorph(this);
				eyesControl = new EyesControl(this);
				belly       = new Belly(this);
				etc         = new Etc(this);
				sound       = new Sound(this);
				unityAssets = new UnityAssets(this);
				addonManager1 = new AddonManager1(this);
				DoAddonManager2Function("Construct");
				worldUi     = new WorldUI(this, 0.001f);
				customUi    = new CustomUI(this);
#if	(EBATWO)				
				orgasm      = new Orgasm(this);
#endif
				InitCore();
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void syncParameter(){
			isDuringOrgasm = false;
			isDuringAfterglow = false;
#if	(EBATWO)
			isDuringOrgasm = orgasm.isDuringOrgasm;
			isDuringAfterglow = orgasm.isDuringAfterglow;
#endif
		}
		
		public void InitCore(){			
			loadSettings("current");
			if(configFileExist() == false){
				loadSettings("default");
				saveConfig();
			}
			applySettings();
			customUi.InitExpandState(false);
			customUi.SetupUIVariables();
			customUi.setupUI();
		}
		
		public void loadSettings(string type){
			try{
				defaultSettingsPath = myPath + "/" + "Config/defaultSettings.json";
				string path = (type == "default") ? defaultSettingsPath : currentSettingsPath;
				settingsRoot = SuperController.singleton.LoadJSON(path) as JSONClass;
				settings = (gender == "Female") ? settingsRoot["Female"].AsObject : settingsRoot["Male"].AsObject;

				if(version != settingsRoot["Common"]["General"]["Version"].AsFloat){
					defaultSettingsPath = myPath + "/" + "Config/defaultSettings.json";
					settingsRoot = SuperController.singleton.LoadJSON(defaultSettingsPath) as JSONClass;					
					settings = (gender == "Female") ? settingsRoot["Female"].AsObject : settingsRoot["Male"].AsObject;
				}
				
				#region general settings from json
				pluginLabel = settingsRoot["Common"]["General"]["Label"];
				disableLog = settingsRoot["Common"]["General"]["Debug Message Off"].AsBool;

				menuOnGlobal = settings["General"]["Display UI Menu"].AsBool;
				
				expression.disableLog = disableLog;
				blushing.disableLog = disableLog;
				tears.disableLog = disableLog;
				bodyMorph.disableLog = disableLog;
				eyesControl.disableLog = disableLog;
				belly.disableLog = disableLog;
				etc.disableLog = disableLog;
				sound.disableLog = disableLog;
				
				stageThreshold = new int[6];
				stageThreshold[0] = 0;
				for(int i = 0; i < 6; i++){
					stageThreshold[i] = settings["StageThreshold"][("Threshold" + i)]["Value"].AsInt;				
				}
				#endregion
				
				#region 
				expression.loadSettings();
				blushing.loadSettings();
				tears.loadSettings();
				bodyMorph.loadSettings();
				eyesControl.loadSettings();
				belly.loadSettings();
				etc.loadSettings();
				sound.loadSettings();
				unityAssets.loadSettings();
				customUi.loadSettings();
				worldUi.loadSettings();

				loadSettingsAlt();
				
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void loadSettingsAlt(){
#if (EBATWO)
			orgasm.loadSettings();
			orgasm.applySettings();
			orgasm.maintenance();
#endif
			addonManager1.loadSettings();
			addonManager1.applySettings();			
			DoAddonManager2Function("LoadSettings");
			DoAddonManager2Function("ApplySettings");
		}

		public void Start(){
			try{
				#region at once: base
				worldUi.makeWorldUI();
				unityAssets.makeUAPee();
				
				openingTime = Time.time;
				logger("openingTime = " + openingTime);
				logger("myPath:" + myPath);
				setGauge(0);
				isBreathActive = false;
				#endregion
				
				maintenanceSequence();
				belly.storePositionAndRotation();
				tears.unequipTearsAll();
				bodyMorph.stopBodyMorph();
				expression.stopExpression();
			}catch (Exception e){
				loggerE(e);
			}			
		}

		public bool configFileExist(){
			try{
				if(FileManagerSecure.FileExists(currentSettingsPath)){
					return true;
				}else{
					return false;
				}
			}catch (Exception e){
				loggerE(e);
				return false;
			}
		}
		
		
		public void applySettings(){
			try{
				pluginLabelJSON.val = settingsRoot["Common"]["General"]["Label"];

				#region
				expression.applySettings();
				blushing.applySettings();
				tears.applySettings();
				bodyMorph.applySettings();
				eyesControl.applySettings();
				sound.applySettings();
				belly.applySettings();
				etc.applySettings();
				unityAssets.applySettings();
				worldUi.applySettings();
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void maintenanceSequence(){
			try{
				#region initialize trigger
				resetTriggers();
				RegisterActions();
				#endregion
				
				#region maintenance
				expression.maintenance();
				blushing.maintenance();
				tears.maintenance();
				bodyMorph.maintenance();
				eyesControl.maintenance();
				belly.maintenance();
				etc.maintenance();
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void DoAddonManager2Function(string func){
#if (ADDON2)
			if(func == "Construct"){
				addonManager2 = new AddonManager2(this);
			}else if(func == "LoadSettings"){
				addonManager2.loadSettings();
			}else if(func == "ApplySettings"){
				addonManager2.applySettings();				
			}else if(func == "Manage"){
				addonManager2.manage(currentTime);				
			}else if(func == "OnDestroy"){
				addonManager2.OnDestroy();				
			}else if(func == "OnOrgasmStart"){
				addonManager2.OnOrgasmStart();				
			}else if(func == "OnOrgasmStart"){
				addonManager2.OnOrgasmStart();				
			}else{
				debug("In EBATCore: Unknown command: " + func);
			}
#endif	
		}
		
		public void saveConfig(){
			try{
				SuperController.singleton.SaveJSON(settingsRoot, currentSettingsPath);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setGauge(int value){
			if(gauge > stageThreshold[5] * 1.2){
				value = (int)(stageThreshold[5] * 1.2);
			}
			gauge = value;
			if(gauge > 0){
				calmingDownTime = currentTime + calmingDownTimeOffset + calmingDownTimeMinusOffset;
			}else{
				calmingDownTime = currentTime + 3600 * 24;
			}
			if(getStage() >= 2){
				isHot = true;
			}else{
				isHot = false;
			}
			worldUi.uiGaugeSlider.slider.value = gauge;
		}
		
		public int getGauge(){
			return gauge;
		}
		
		public void incrementGauge(int value = 1){
			setGauge(gauge + value);
		}
		
		public void decrementGauge(){
			if(isDuringOrgasm){
				return;
			}
			setGauge(gauge - 1 - calmingDownPlus);
		}
		
		#region trigger functions
		public void RegisterActions(){
			try{
//				RegisterActionPair("LipTrigger", "OnLipTouch", "Lip");				
				RegisterActionPair("MouthTrigger", "OnMouthTouch", "Mouth");				
				RegisterActionPair("ThroatTrigger", "OnThroatTouch", "Throat");				
				RegisterActionPair("lNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterActionPair("rNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterActionPair("LabiaTrigger", "OnLabiaTouch", "Labia");
				RegisterActionPair("VaginaTrigger", "OnVaginaTouch", "Vagina");
				RegisterActionPair("DeepVaginaTrigger", "OnVaginaDeepTouch", "VaginaDeep");
				RegisterActionPair("DeeperVaginaTrigger", "OnVaginaDeeperTouch", "VaginaDeeper");
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void RegisterActionPair(string collisionTrigger, string action, string position){
			try{
				JSONStorableAction tmpAction = new JSONStorableAction(action, () =>{
					OnErogenousZoneTouch(position);
				});
				RegisterAction(tmpAction);
				createTrigger(collisionTrigger, action);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void resetTriggers() {
			try{
				var table = new Dictionary<string, int>{
//					{"OnLipTouch", 1},
					{"OnMouthTouch", 1},
					{"OnThroatTouch", 1},
					{"OnNippleTouch", 1},
					{"OnLabiaTouch", 1},
					{"OnVaginaTouch", 1},
					{"OnVaginaDeepTouch", 1},
					{"OnVaginaDeeperTouch", 1}
				};
//				string[] cTriggerNames = {"LipTrigger", "MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				foreach(string cTriggerName in cTriggerNames){
					logger("cTriggerName = " + cTriggerName);
					CollisionTrigger cTrigger = atom.GetStorableByID(cTriggerName) as CollisionTrigger;
					JSONClass triggerClass = cTrigger.trigger.GetJSON();
					JSONArray triggerArray = triggerClass["startActions"].AsArray;
					for (int i=0;i< triggerArray.Count;i++) {
						logger("triggerArray[" + i + "][name]=" + triggerArray[i]["name"].Value);
						if (table.ContainsKey(triggerArray[i]["name"].Value)){
							triggerArray.Remove(i);
						}
					}
					cTrigger.trigger.RestoreFromJSON(triggerClass);
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void createTrigger(string triggerCollisionTriggerName, string scriptTriggerName){
			try{
				CollisionTrigger cTrigger = atom.GetStorableByID(triggerCollisionTriggerName) as CollisionTrigger;
				cTrigger.enabled = true;
				TriggerActionDiscrete startTrigger;
				startTrigger=cTrigger.trigger.CreateDiscreteActionStartInternal();
				startTrigger.name = scriptTriggerName;
				startTrigger.receiverAtom = atom;
				startTrigger.receiver = GetPluginStorableById(atom, myClassName);
				startTrigger.receiverTargetName = scriptTriggerName;
			}catch (Exception e){
				loggerE(e);
			}
		}
	
		// Function taken from VAMDeluxe's code
		static JSONStorable GetPluginStorableById(Atom atom, string id){
			return Utility.GetPluginStorableById(atom, id);
		}
		#endregion		
		
		#region engine
		public void OnErogenousZoneTouch(string zone){ 
			try{
				logger("Touch: " + zone);
				currentTime = Time.time;
				lastTouchingTime = currentTime;
				sound.playSound("moan", Time.time);
				updateAll();			

				calmingDownPlus = 0;
				calmingDownTimeMinusOffset = 0;
				int value = calcZoneMap(zone);
				incrementGauge(value);
				isStageChanged();
				
				if(isBreathActive == false){
					isBreathActive = true;
				}
				
				orgasmFunctions("orgasmCheck", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public int calcZoneMap(string zone){
			int stage = getStage();
			int value = 0;
			if(zone == "Mouth" || zone == "Throat"){
				if(stage <= 2){
					value = 3;
				}else if(stage <= 4){
					value = 2;
				}else{
					value = 1;
				}
			}else if(zone == "Nipple"){
				if(stage <= 2){
					value = 2;
				}else if(stage <= 4){
					value = 4;
				}else{
					value = 1;
				}
			}else if(zone == "Labia" || zone == "Vagina"){
				if(stage <= 2){
					value = 1;
				}else if(stage <= 4){
					value = 3;
				}else{
					value = 4;
				}
			}else if(zone == "VaginaDeep" || zone == "VaginaDeeper"){
				if(stage <= 2){
					value = 1;
				}else if(stage <= 4){
					value = 2;
				}else{
					value = 6;
				}
			}else if(zone == "ByScript"){
				value = 1;
			}
			return value;
		}
		
		public void OnStageChange(){
			try{
				tears.cm.unequipClothes(preStage);
				tears.cm.equipClothes(curStage);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void isStageChanged(){
			try{
				if(curStage != getStage()){
					preStage = curStage;
					curStage = getStage();
					logger("stage changed: " + preStage + " => " + curStage);
					OnStageChange();
				}
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void orgasmFunctions(string func, float time){
#if (EBATWO)
			if(func == "orgasmCheck"){
				orgasm.orgasmCheck(time);
			}else if(func == "update"){
				orgasm.update(time);
			}else if(func == "orgasmControl"){
				orgasm.orgasmControl(time);
			}else if(func == "orgasmStart"){
				orgasm.orgasmStart();
			}else if(func == "orgasmFinish"){
				orgasm.orgasmFinish();
			}
#endif
		}
		
		public bool isOrgasmFunctionEnabled(){
			bool flag = false;
#if (EBATWO)
			flag = true;
#endif
			return flag;
		}
		
		public void updateAll(){
			try{
				expression.update(currentTime);
				blushing.update(currentTime);
				tears.update(currentTime);
				bodyMorph.update(currentTime);
				eyesControl.update(currentTime);
				orgasmFunctions("update", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public float getTime(){
			return Time.time;
		}
		
        public void Update(){
            try{
				if(Time.time > currentTime + 10){
					currentTime = Time.time;
					logger("start plug-in or load scene.");
					maintenanceSequence();
				}
				
				string currentName = atom.name;
				if(myName != currentName){
					logger("old name:" + myName + ", new name:" + currentName);
					myName = currentName;
					maintenanceAlert = true;
				}
				
				if(maintenanceAlert){
					maintenanceAlert = false;
					Init();
					customUi.RemoveConfigMenu();
					Start();
					return;
				}
				
				currentTime = Time.time;
				PerSecondCheck();

				if(calmingDownTime < currentTime){
					decrementGauge();
				}
				
				orgasmFunctions("orgasmControl", currentTime);
				blushing.manage(currentTime);
				tears.manage(currentTime);
				bodyMorph.manage(currentTime);
				expression.manage(currentTime);
				eyesControl.manage(currentTime);
				belly.manage(currentTime);
				unityAssets.manage(currentTime);
				addonManager1.manage(currentTime);
				DoAddonManager2Function("Manage");
				
				if(expression.isExpressionActive == false && blushing.isBlushingActive == false && tears.isTearsActive == false){
					if(isDuringOrgasm == false){
						setGauge(0);
						isStageChanged();
					}
				}
				
				if(isOrgasmFunctionEnabled() == false &&
					lastTouchingTime + longestDurationTime() < currentTime){
					setGauge(0);
					isStageChanged();
				}
			}catch (Exception e){
//				loggerE(e);
			}
		}

		public void PerSecondCheck(){
			if(lastTime + perSecondOffset < currentTime){
				syncParameter();
				lastTime = currentTime;
				customUi.MemorizeUIPosition();
			}
		}
		
		public float longestDurationTime(){
			float eTime = expression.expressionClosingTimeOffset1 + expression.expressionClosingTimeOffset2;
			float bTime = blushing.blushClosingTimeOffset;
			float tTime = tears.tearsClosingTimeOffset;
			float tmpTime = (eTime > bTime) ? eTime : bTime;
			float result = (tmpTime > tTime) ? tmpTime : tTime;
			return result;
		}
		
		public int getStage(){
			try{
				int stage = 0;
				int gauge = getGauge();
				if(gauge < stageThreshold[1]){
					stage = 0;
				}else if(gauge < stageThreshold[2]){
					stage = 1;
				}else if(gauge < stageThreshold[3]){
					stage = 2;
				}else if(gauge < stageThreshold[4]){
					stage = 3;
				}else if(gauge < stageThreshold[5]){
					stage = 4;
				}else{
					stage = 5;
				}
				return stage;
			}catch (Exception e){
				loggerE(e);
				return -1;
			}
		}
		
		#endregion
		
		#region etc
		public void logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		public void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public virtual IEnumerator CreateAtom(string atomType, string prefix, Action<Atom> onAtomCreated, bool keep=false)
        {
            //  need to wait a bit before creating if we are just loading into the scene, as the previous atom may have not been fully destroyed!!
//            yield return new WaitForSeconds(1.0f);
            yield return new WaitForSeconds(0.1f);

            string atomId = GenerateAtomName(prefix, atom.uid);
            Atom tmpatom = SuperController.singleton.GetAtomByUid(atomId);
            if (tmpatom == null)
            {
                yield return SuperController.singleton.AddAtomByType(atomType, atomId);
                tmpatom = SuperController.singleton.GetAtomByUid(atomId);
            }

            if (tmpatom != null)
            {
                if (keep == false)
                {
                    generatedAtoms.Add(tmpatom);
					generatedAtomsDictionary.Add(atomId, tmpatom);
                }
                onAtomCreated(tmpatom);
            }
        }

		public string GenerateAtomName(string prefix, string suffix)
        {
            return prefix + " " + suffix;
        }

		#endregion
		
		public void OnDestroy(){
			finishSequence();
			worldUi.OnDestroy();
			unityAssets.OnDestroy();
			
			expression.OnDestroy();
			tears.OnDestroy();
			blushing.OnDestroy();
			bodyMorph.OnDestroy();
			sound.OnDestroy();
			belly.OnDestroy();
			addonManager1.OnDestroy();
			DoAddonManager2Function("OnDestroy");

            generatedAtoms.ForEach((atom) =>
            {
                if (atom == null)
                {
                    return;
                }
                SuperController.singleton.RemoveAtom(atom);
            });
            generatedAtoms.Clear();			
		}
		
		public void OnOrgasmStart()
		{
			worldUi.OnOrgasmStart();
			addonManager1.OnOrgasmStart();
			DoAddonManager2Function("OnOrgasmStart");
		}
		
		public void OnOrgasmStop()
		{
			worldUi.OnOrgasmStop();
			addonManager1.OnOrgasmStop();
			DoAddonManager2Function("OnOrgasmStop");
		}
		
		public void finishSequence(){
			expression.stopExpression();
			tears.unequipTearsAll();
			blushing.RemoveBlushing();
			bodyMorph.stopBodyMorph();			
		}
		
		public void surveyStorables(int lvl){
			List<string> ids = new List<string>();
			ids = atom.GetStorableIDs();
			for(int i = 0; i < ids.Count; i++){
				SuperController.LogMessage("id: " + ids[i]);
				List<string> names = new List<string>();
				
				if(lvl > 0){
					names = atom.GetStorableByID(ids[i]).GetBoolParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  bool: " + names[j]);
					}
				}
					
				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllParamAndActionNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  ParamAndAction: " + names[j]);
					}					
				}

				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllFloatAndColorParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  FloatAndColorParam: " + names[j]);
					}
				}
			}
		}

	}
}
